/*
  Copyright 2021 Kakao Corp.

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */
package com.kakao.sdk.partner.friend

import android.app.Activity
import android.content.Context
import android.content.Intent
import android.os.*
import com.kakao.sdk.common.model.ClientError
import com.kakao.sdk.common.model.ClientErrorCause
import com.kakao.sdk.common.model.KakaoSdkError
import com.kakao.sdk.common.util.SdkLog
import com.kakao.sdk.partner.friend.model.*
import com.kakao.sdk.partner.friend.network.model.PickerChatFilter
import com.kakao.sdk.partner.friend.util.ChatProfileImageMaker
import com.kakao.sdk.partner.friend.view.*
import java.util.concurrent.atomic.AtomicInteger

/**
 * @suppress
 */
internal object PickerManager {
    @JvmStatic
    fun startActivity(
        isSingle: Boolean = false,
        isPopup: Boolean = false,
        context: Context, param: PickerFriendRequestParams,
        callback: (selectedUsers: SelectedUsers?, error: Throwable?) -> Unit
    ) {
        if (!isSingle && !isParameterCorrect(param, callback)) {
            return
        }

        val requestParam = param as PickerFriendRequestParams
        Repository.loadFriendsData(
            requestParam.pickerFriendOrder,
            requestParam.pickerFriendType ?: Constants.DEFAULT_PICKER_FRIEND_TYPE,
            requestParam.pickerFriendFilter ?: Constants.DEFAULT_PICKER_FRIEND_FILTER,
            requestParam.countryCodes
        ) { error ->
            if (error != null) {
                Repository.clearData()
                callback(null, error)
            } else {
                val activity =
                    if (isPopup) FriendPopupPickerActivity::class.java else FriendPickerActivity::class.java
                val intent =
                    Intent(context, activity)
                        .putExtra(Constants.KEY_BUNDLE, Bundle().apply {
                            putBoolean(Constants.KEY_IS_SINGLE, isSingle)
                            putBoolean(Constants.KEY_IS_POPUP, isPopup)
                            putParcelable(
                                Constants.KEY_RESULT_RECEIVER,
                                resultReceiver<SelectedUser, SelectedUsers>(
                                    "***** Friend Picker RESULT:",
                                    true,
                                    Constants.KEY_SELECTED_FRIENDS,
                                    callback
                                )
                            )
                            putParcelable(Constants.KEY_FRIEND_PICKER_PARAM, param)
                        })
                        .addFlags(Intent.FLAG_ACTIVITY_NEW_TASK)
                context.startActivity(intent)
            }
        }
    }

    @JvmStatic
    fun startActivityForChat(
        isPopup: Boolean = false,
        context: Context, param: PickerChatRequestParams,
        callback: (selectedUsers: SelectedUsers?, error: Throwable?) -> Unit
    ) {
        startActivityForChat(
            isPopup,
            false,
            context,
            param,
            "***** Chat Member Picker RESULT:",
            resultKey = Constants.KEY_SELECTED_FRIENDS,
            callback = callback
        )
    }

    // TODO: param 받는 방식이 수정될 수 있음
    @JvmStatic
    fun startActivity(
        isSingle: Boolean = false,
        isPopup: Boolean = false,
        context: Context,
        friendParam: PickerFriendRequestParams,
        chatParam: PickerChatRequestParams,
        callback: (selectedUsers: SelectedUsers?, error: Throwable?) -> Unit
    ) {
        if (!isParameterCorrect(friendParam, callback) && !isParameterCorrect(
                chatParam,
                callback
            )
        ) {
            return
        }

        val count = AtomicInteger()

        Repository.loadFriendsData(
            friendParam.pickerFriendOrder,
            friendParam.pickerFriendType ?: Constants.DEFAULT_PICKER_FRIEND_TYPE,
            friendParam.pickerFriendFilter ?: Constants.DEFAULT_PICKER_FRIEND_FILTER,
            friendParam.countryCodes
        ) { error ->
            count.incrementAndGet()
            if (count.get() == 2) {
                if (error != null) {
                    Repository.clearData()
                    callback(null, error)
                } else {
                    count.incrementAndGet()

                    startActivityForTab(
                        context,
                        isSingle,
                        isPopup,
                        friendParam,
                        chatParam,
                        callback
                    )
                }
            }
        }

        // TODO: 챗 멤버 피커일 때는 오픈 카톡 안보이도록 처리해야함. 챗 피커, 챗 멤버 피커 구분 방식에 따라 구현이 조금 달라질 수 있음
        Repository.loadChatsData(filters = listOf(PickerChatFilter.REGULAR)) { error ->
            count.incrementAndGet()
            if (count.get() == 2) {
                if (error != null) {
                    Repository.clearData()
                    callback(null, error)
                } else {
                    startActivityForTab(
                        context,
                        isSingle,
                        isPopup,
                        friendParam,
                        chatParam,
                        callback
                    )
                }
            }
        }
    }

    @JvmStatic
    fun startActivityOnlyChat(
        isPopup: Boolean = false,
        context: Context, param: PickerChatRequestParams,
        callback: (selectedChat: SelectedChat?, error: Throwable?) -> Unit
    ) {
        startActivityForChat(
            isPopup,
            true,
            context,
            param,
            "***** Chat Picker RESULT:",
            resultKey = Constants.KEY_SELECTED_CHAT,
            callback = callback
        )
    }

    @JvmStatic
    private fun startActivityForTab(
        context: Context,
        isSingle: Boolean,
        isPopup: Boolean,
        friendParam: PickerFriendRequestParams,
        chatParam: PickerChatRequestParams,
        callback: (selectedUsers: SelectedUsers?, error: Throwable?) -> Unit
    ) {
        val activity =
            if (isPopup) TabPopupPickerActivity::class.java else TabPickerActivity::class.java
        val intent =
            Intent(context, activity).putExtra(Constants.KEY_BUNDLE, Bundle().apply {
                putBoolean(Constants.KEY_IS_SINGLE, isSingle)
                putBoolean(Constants.KEY_IS_POPUP, isPopup)
                putBoolean(Constants.KEY_IS_TAB, true)
                putBoolean(Constants.KEY_IS_CHAT_MEMBER_PICKER, false)
                putParcelable(
                    Constants.KEY_RESULT_RECEIVER,
                    resultReceiver<SelectedUser, SelectedUsers>(
                        "***** Friend Picker RESULT:",
                        true,
                        Constants.KEY_SELECTED_FRIENDS,
                        callback
                    )
                )
                putParcelable(Constants.KEY_FRIEND_PICKER_PARAM, friendParam)
                putParcelable(Constants.KEY_CHAT_PICKER_PARAM, chatParam)
            })
                .addFlags(Intent.FLAG_ACTIVITY_NEW_TASK)
        context.startActivity(intent)
    }

    @JvmStatic
    private inline fun <reified T : Parcelable, reified R : Parcelable> startActivityForChat(
        isPopup: Boolean,
        isChatOnly: Boolean,
        context: Context,
        param: T,
        message: String,
        isFriendPicker: Boolean = false,
        resultKey: String,
        crossinline callback: (result: R?, error: Throwable?) -> Unit
    ) {
        val filter = if (!isChatOnly) listOf(PickerChatFilter.REGULAR) else null
        Repository.loadChatsData(filter) { error ->
            if (error != null) {
                Repository.clearData()
                callback(null, error)
            } else {
                val activity =
                    if (isPopup) ChatPopupPickerActivity::class.java else ChatPickerActivity::class.java
                val intent =
                    Intent(context, activity).putExtra(Constants.KEY_BUNDLE, Bundle().apply {
                        putBoolean(Constants.KEY_IS_POPUP, isPopup)
                        putBoolean(Constants.KEY_IS_CHAT_ONLY, isChatOnly)
                        putParcelable(
                            Constants.KEY_RESULT_RECEIVER,
                            resultReceiver<Parcelable, R>(
                                message,
                                isFriendPicker,
                                resultKey,
                                callback
                            )
                        )
                        putParcelable(Constants.KEY_CHAT_PICKER_PARAM, param)
                    })
                        .addFlags(Intent.FLAG_ACTIVITY_NEW_TASK)
                context.startActivity(intent)
            }
        }
    }

    @JvmSynthetic
    private inline fun <reified T : Parcelable, reified R : Parcelable> resultReceiver(
        message: String,
        isArray: Boolean = false,
        resultKey: String,
        crossinline callback: (results: R?, error: Throwable?) -> Unit
    ): ResultReceiver = object : ResultReceiver(Handler(Looper.getMainLooper())) {
        override fun onReceiveResult(resultCode: Int, resultData: Bundle?) {
            SdkLog.d("$message $resultData")
            when (resultCode) {
                Activity.RESULT_OK -> {
                    var result: R? = null
                    if (isArray) {
                        resultData?.getParcelableArrayList<T>(resultKey)?.let {
                            result = R::class.java.getConstructor(List::class.java).newInstance(it)
                        }
                    } else {
                        resultData?.getParcelable<R>(resultKey)?.let {
                            result = it
                        }
                    }
                    callback(result, null)
                }
                Activity.RESULT_CANCELED -> {
                    val error =
                        resultData?.getSerializable(Constants.KEY_EXCEPTION) as KakaoSdkError
                    callback(null, error)
                }
                else -> {
                    val error = IllegalArgumentException("Unknown resultCode")
                    callback(null, error)
                }
            }
            ChatProfileImageMaker.clearCacheImage() // Chat Profile Image 제거
            Repository.clearData() // Repository 클래스에 있는 데이터 제거
        }
    }

    @JvmStatic
    private fun isParameterCorrect(
        param: Parcelable,
        callback: (selectedUsers: SelectedUsers?, error: Throwable?) -> Unit
    ): Boolean {
        val minPickableCount =
            if (param is PickerFriendRequestParams) param.minPickableCount
                ?: Constants.DEFAULT_MIN_PICKABLE_COUNT
            else (param as PickerChatRequestParams).minPickableCount
                ?: Constants.DEFAULT_MIN_PICKABLE_COUNT
        val maxPickableCount =
            if (param is PickerFriendRequestParams) param.maxPickableCount
                ?: Constants.DEFAULT_MAX_PICKABLE_COUNT
            else (param as PickerChatRequestParams).maxPickableCount
                ?: Constants.DEFAULT_MAX_PICKABLE_COUNT

        val exception = when {
            minPickableCount < Constants.MULTI_PICKER_MIN_PICKABLE_COUNT ->
                ClientError(
                    ClientErrorCause.BadParameter,
                    "Parameter minPickableCount must be greater than 1."
                )

            maxPickableCount > Constants.MULTI_PICKER_MAX_PICKABLE_COUNT ->
                ClientError(
                    ClientErrorCause.BadParameter,
                    "Parameter maxPickableCount must be 100 or less."
                )

            minPickableCount > maxPickableCount ->
                ClientError(
                    ClientErrorCause.BadParameter,
                    "Parameter maxPickableCount is greater than or equal to parameter minPickableCount"
                )

            else -> null
        }

        exception?.let {
            resultReceiver<SelectedUser, SelectedUsers>(
                "***** Friend Picker RESULT:",
                true,
                Constants.KEY_SELECTED_FRIENDS,
                callback
            ).send(Activity.RESULT_CANCELED, Bundle().apply {
                putSerializable(Constants.KEY_EXCEPTION, it)
            })
            return false
        }
        return true
    }
}