/*
  Copyright 2021 Kakao Corp.

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */
package com.kakao.sdk.partner.friend.adapter

import android.content.res.Configuration
import android.view.LayoutInflater
import android.view.View
import android.view.ViewGroup
import android.widget.LinearLayout
import androidx.constraintlayout.widget.ConstraintSet
import androidx.constraintlayout.widget.ConstraintSet.*
import androidx.recyclerview.widget.DiffUtil
import androidx.recyclerview.widget.RecyclerView
import com.kakao.sdk.partner.friend.imageloader.ImageLoader
import com.kakao.sdk.partner.friend.model.ChatItem
import com.kakao.sdk.partner.friend.model.FriendItem
import com.kakao.sdk.partner.friend.model.Item
import com.kakao.sdk.partner.friend.util.LayoutUtils
import com.kakao.sdk.partner.friend.util.getDefaultProfileImage
import com.kakao.sdk.v2.partner.friend.R
import com.kakao.sdk.v2.partner.friend.databinding.*

/**
 * @suppress
 */
internal class FriendsListAdapter(
    private val isPopup: Boolean,
    private val isSingle: Boolean,
    private val enableSearch: Boolean,
    private val showPickedFriend: Boolean,
    private val maxPickableCount: Int,
    private val checkCallback: (FriendItem.User?) -> Unit,
    private val uncheckCallback: (FriendItem.User) -> Unit,
    private var isChatMemeberPicker: Boolean = false,
    private var totalFriendsCount: Int,
    private val selectAllCallback: (Boolean) -> Unit
) : RecyclerView.Adapter<RecyclerView.ViewHolder>() {
    private var items = mutableListOf<Item>() // 화면에 보여지는 전체 아이템 리스트
    private var pickedFriends = listOf<FriendItem.User>() // 친구 추가 선택 가능 여부 파악을 위한 리스트
    private lateinit var mRecyclerView: RecyclerView // 스크롤 처리를 위한 recyclerview
    private val diffUtilCallback = ItemDiffUtilCallback()

    override fun onCreateViewHolder(parent: ViewGroup, viewType: Int): RecyclerView.ViewHolder {
        val inflater = LayoutInflater.from(parent.context)
        return when (viewType) {
            VIEW_TYPE_HEADER -> {
                HeaderViewHolder(KakaoSdkHeaderFriendPickerBinding.inflate(inflater, parent, false))
            }
            VIEW_TYPE_USER -> {
                UserViewHolder(isPopup,
                    KakaoSdkItemFriendPickerBinding.inflate(inflater, parent, false),
                    checkCallback = {
                        // 최대 선택 가능한 친구를 다 채운 경우
                        if (pickedFriends.size == maxPickableCount) {
                            // 싱글 피커인 경우 이전에 선택된 친구 삭제 후 새로운 친구 리턴
                            if (isSingle) {
                                uncheckCallback(pickedFriends[0])
                                checkCallback(it)
                            } else {
                                // 멀티 피커이면서 이미 최대 선택 가능한 친구를 다 채운 경우 null 리턴
                                checkCallback(null)
                            }
                        } else {
                            checkCallback(it)
                        }
                    },
                    uncheckCallback = {
                        uncheckCallback(it)
                    })
            }
            VIEW_CATEGORY_DIVIDER -> {
                DividerViewHolder(KakaoSdkCategoryDividerBinding.inflate(inflater, parent, false))
            }
            VIEW_EMPTY_SEARCH_RESULT_TEXT -> {
                EmptySearchViewHolder(
                    isPopup,
                    KakaoSdkEmptySearchResultTextBinding.inflate(inflater, parent, false)
                )
            }
            VIEW_SELECT_ALL_BUTTON -> {
                SelectAllViewHolder(KakaoSdkItemSelectAllBinding.inflate(inflater, parent, false)) {
                    selectAllCallback(it)
                }
            }
            else -> throw ClassCastException("Unknown viewType $viewType")
        }
    }

    override fun onBindViewHolder(holder: RecyclerView.ViewHolder, position: Int) {
        val item = items[position]
        when (holder) {
            is HeaderViewHolder -> holder.bind(item as Item.Header)
            is UserViewHolder -> holder.bind(item as FriendItem.User, isSingle)
            is EmptySearchViewHolder -> holder.setLayout()
            is SelectAllViewHolder -> holder.bind(item as ChatItem.SelectAll)
        }
    }

    override fun getItemCount(): Int = items.size

    override fun getItemViewType(position: Int): Int {
        return when (items[position]) {
            is Item.Header -> VIEW_TYPE_HEADER
            is FriendItem.User -> VIEW_TYPE_USER
            is Item.Divider -> VIEW_CATEGORY_DIVIDER
            is Item.EmptySearchResult -> VIEW_EMPTY_SEARCH_RESULT_TEXT
            is ChatItem.SelectAll -> VIEW_SELECT_ALL_BUTTON
            else -> throw ClassCastException("Unknown viewType ${items[position]::class.java}")
        }
    }

    override fun onAttachedToRecyclerView(recyclerView: RecyclerView) {
        super.onAttachedToRecyclerView(recyclerView)
        mRecyclerView = recyclerView
    }

    fun setFriendsList(items: MutableList<Item>) {
        diffUtilCallback.setLists(this.items, items)
        DiffUtil.calculateDiff(diffUtilCallback).dispatchUpdatesTo(this)
        this.items = items
    }

    fun checkFriend(friend: FriendItem.User) {
        if (!isSingle && !enableSearch && showPickedFriend && pickedFriends.isEmpty()) {
            addItemDividerAtTop()
        }
        updateDuplicatedFriend(friend, true)
    }

    fun uncheckFriend(friend: FriendItem.User) {
        if (!isSingle && !enableSearch && showPickedFriend && pickedFriends.size == 1) {
            items.removeAt(0)
            notifyItemRemoved(0)
        }
        updateDuplicatedFriend(friend, false)
    }

    fun setPickedFriends(pickedFriends: MutableList<FriendItem.User>) {
        this.pickedFriends = pickedFriends

        if (isChatMemeberPicker) {
            setSelectAllButton()
        }
    }

    fun clearPickedFriends() {
        pickedFriends.forEach { uncheckFriend(it) }
    }

    fun checkAllFriends() {
        if (!enableSearch && items[0] !is Item.Divider) {
            addItemDividerAtTop()
        }
        items.forEach { if (it is FriendItem.User) it.isChecked = true }
        pickedFriends =
            items.filterIsInstance<FriendItem.User>().filter { !pickedFriends.contains(it) }
                .toList()
        notifyDataSetChanged()
    }

    fun uncheckAllFriends() {
        if (!enableSearch && items[0] is Item.Divider) {
            items.removeAt(0)
            notifyItemRemoved(0)
        }
        items.forEach { if (it is FriendItem.User) it.isChecked = false }
        pickedFriends = listOf()
        notifyDataSetChanged()
    }

    private fun setSelectAllButton() {
        if (pickedFriends.size == totalFriendsCount) {
            val index = findSelectAllButton()
            (items[index] as ChatItem.SelectAll).isChecked = true
            notifyItemChanged(index)
        } else if (isChatMemeberPicker && pickedFriends.size + 1 == totalFriendsCount) {
            val index = findSelectAllButton()
            (items[index] as ChatItem.SelectAll).isChecked = false
            notifyItemChanged(index)
        }
    }

    private fun findSelectAllButton(): Int {
        var index = 0
        for (i in 0 until items.size) {
            val item = items[i]
            if (item is ChatItem.SelectAll) {
                index = i
                break
            }
        }
        return index
    }

    private fun addItemDividerAtTop() {
        items.add(0, Item.Divider)
        notifyItemInserted(0)
        // 구분선 추가 전 화면이 리스트 최상단이면 새로 추가된 구분선까지 스크롤을 위로 올려줌.
        if (mRecyclerView.computeVerticalScrollOffset() == 0) {
            mRecyclerView.scrollToPosition(0)
        }
    }

    private fun updateDuplicatedFriend(user: FriendItem.User, isChecked: Boolean) {
        val indexOfFirst = items.indexOfFirst { it == user }
        val indexOfLast = items.indexOfLast { it == user }
        // index를 못찾은 경우에는 -1 이므로 crash 방지 (검색 중에는 리스트에 해당 user가 없을 수 있음)
        if (indexOfFirst > -1) {
            (items[indexOfFirst] as FriendItem.User).isChecked = isChecked
        }
        if (indexOfLast > -1) {
            (items[indexOfLast] as FriendItem.User).isChecked = isChecked
        }
        notifyItemChanged(indexOfFirst)
        notifyItemChanged(indexOfLast)
    }

    class HeaderViewHolder(
        private val binding: KakaoSdkHeaderFriendPickerBinding
    ) : RecyclerView.ViewHolder(binding.root) {
        fun bind(header: Item.Header) {
            binding.titleTv.text = header.title
        }
    }

    class DividerViewHolder(binding: KakaoSdkCategoryDividerBinding) :
        RecyclerView.ViewHolder(binding.root)

    class EmptySearchViewHolder(
        private val isPopup: Boolean,
        private val binding: KakaoSdkEmptySearchResultTextBinding
    ) : RecyclerView.ViewHolder(binding.root) {
        fun setLayout() {
            binding.apply {
                if (LayoutUtils.orientation == Configuration.ORIENTATION_PORTRAIT) {
                    if (isPopup) {
                        emptySearchResultText.apply {
                            (layoutParams as LinearLayout.LayoutParams).topMargin =
                                resources.getDimensionPixelSize(R.dimen.popup_empty_search_result_margin_top)
                        }
                    }
                } else {
                    emptySearchResultText.apply {
                        (layoutParams as LinearLayout.LayoutParams).topMargin = if (isPopup) {
                            resources.getDimensionPixelSize(R.dimen.popup_empty_search_result_margin_top_landscape)
                        } else {
                            resources.getDimensionPixelSize(R.dimen.empty_search_result_margin_top_landscape)
                        }
                    }
                }
            }
        }
    }

    class SelectAllViewHolder(
        private val binding: KakaoSdkItemSelectAllBinding,
        private var selectAllCallback: (Boolean) -> Unit
    ) : RecyclerView.ViewHolder(binding.root) {

        fun bind(selectAllButton: ChatItem.SelectAll) {
            setCheckAllButton(selectAllButton.isChecked)

            binding.root.setOnClickListener {
                selectAllButton.isChecked = !selectAllButton.isChecked
                setCheckAllButton(selectAllButton.isChecked)
                selectAllCallback(selectAllButton.isChecked)
            }
        }

        private fun setCheckAllButton(isAllChecked: Boolean) {
            binding.selectAllText.text = if (isAllChecked) "전체 해제" else "전체 선택"
            val drawable =
                if (isAllChecked) R.drawable.ic_baseline_check_24_yellow else R.drawable.ic_baseline_check_24_grey
            binding.checkImage.setImageResource(drawable)
            binding.checkBox.isChecked = isAllChecked
        }
    }

    class UserViewHolder(
        private val isPopup: Boolean,
        private val binding: KakaoSdkItemFriendPickerBinding,
        private val checkCallback: (FriendItem.User) -> Unit,
        private val uncheckCallback: (FriendItem.User) -> Unit
    ) : RecyclerView.ViewHolder(binding.root) {

        fun bind(user: FriendItem.User, isSingle: Boolean) {
            binding.apply {
                if (user.disableSelectReasons == null) {
                    // 수신 거부 상태가 아닐 때
                    checkBox.visibility = View.VISIBLE
                    setAlpha(1f)
                    root.setOnClickListener {
                        if (!checkBox.isChecked) {
                            checkCallback(user)
                        } else {
                            uncheckCallback(user)
                        }
                    }
                } else {
                    // 수신 거부 상태일 때
                    checkBox.visibility = View.GONE
                    setAlpha(0.3f)
                    root.setOnClickListener(null) // 선택 불가능하도록 처리
                }

                setProfileImageView(user)
                setNicknameTextView(isPopup, user)
                setRejectTextView(user)
                setCheckbox(isSingle, user)
            }
        }

        private fun setCheckbox(isSingle: Boolean, user: FriendItem.User) {
            binding.checkBox.apply {
                val checkBoxDrawable = if (isSingle) {
                    R.drawable.kakao_sdk_daynight_radio
                } else {
                    R.drawable.kakao_sdk_daynight_checkbox
                }

                setButtonDrawable(checkBoxDrawable)
                isChecked = user.isChecked
            }
        }

        private fun setProfileImageView(user: FriendItem.User) {
            binding.apply {
                val defaultProfileImage =
                    getDefaultProfileImage(user.uuid.hashCode(), root.context).resourceId
                ImageLoader.get(itemView.context).load(user.profileImageUrl)
                    .placeHolder(defaultProfileImage).into(profileIv)
            }
        }

        private fun setNicknameTextView(isPopup: Boolean, user: FriendItem.User) {
            binding.nicknameTv.apply {
                val topMargin = if (isPopup) {
                    if (user.disableSelectReasons == null) resources.getDimensionPixelSize(R.dimen.popup_list_item_nickname_margin_top)
                    else resources.getDimensionPixelSize(R.dimen.popup_list_item_nickname_reject_margin_top)
                } else {
                    if (user.disableSelectReasons == null) resources.getDimensionPixelSize(R.dimen.list_item_nickname_margin_top)
                    else resources.getDimensionPixelSize(R.dimen.list_item_nickname_reject_margin_top)
                }
                val bottomMargin = if (isPopup) {
                    if (user.disableSelectReasons == null) resources.getDimensionPixelSize(R.dimen.popup_list_item_nickname_margin_bottom)
                    else resources.getDimensionPixelSize(R.dimen.popup_list_item_nickname_reject_margin_bottom)
                } else {
                    if (user.disableSelectReasons == null) resources.getDimensionPixelSize(R.dimen.list_item_nickname_margin_bottom)
                    else resources.getDimensionPixelSize(R.dimen.list_item_nickname_reject_margin_bottom)
                }
                text = user.nickname
                ConstraintSet().apply {
                    clone(binding.root)
                    connect(id, TOP, PARENT_ID, TOP, topMargin)
                    connect(id, BOTTOM, PARENT_ID, BOTTOM, bottomMargin)
                }.applyTo(binding.root)
            }
        }

        private fun setRejectTextView(user: FriendItem.User) {
            binding.rejectTv.visibility =
                if (user.disableSelectReasons == null) View.GONE else View.VISIBLE
        }

        private fun setAlpha(value: Float) {
            binding.apply {
                profileIv.alpha = value
                nicknameTv.alpha = value
                rejectTv.alpha = value
            }
        }
    }

    companion object {
        const val VIEW_TYPE_HEADER = 0
        const val VIEW_TYPE_USER = 1
        const val VIEW_CATEGORY_DIVIDER = 2
        const val VIEW_EMPTY_SEARCH_RESULT_TEXT = 3
        const val VIEW_SELECT_ALL_BUTTON = 4
    }
}